#region File Comment
/*
 * Autor: Alexander Grfenstein
 * Date: 2008.29.03
 * Filename: TextField.cs
 */
#endregion

#region Using Namespaces
using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Graphics;
using CellarDoor.Utilities;
using CellarDoor.Components;
#endregion

namespace CellarDoor.Gui
{
    /// <summary>
    /// 
    /// </summary>
    public class UiTextField : UiControl
    {
        #region Variables
        string cursor = "|";
        string text;
        bool cursorIsVisible;
        int cursorBlinkrate = 250;
        int cursorBlinkTimer;
        Vector2 cursorPosition;
        TextEditor editor;
        bool multiline;
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the cursorblinktime in miliseconds.
        /// Default is 250ms;
        /// </summary>
        public int CursorBlinkRate
        {
            get { return cursorBlinkrate; }
            set { cursorBlinkrate = value; }
        }

        /// <summary>
        /// Gets or sets the character that represents the cursor.
        /// Default is '|'
        /// </summary>
        public char Cursor
        {
            get { return cursor.ToCharArray()[0]; }
            set { cursor = value.ToString(); }
        }

        public bool Multiline
        {
            get { return multiline; }
            set { multiline = value; }
        }
        #endregion

        #region Constructor
        /// <summary>
        /// Creates a textfield.
        /// </summary>
        /// <param name="font"></param>
        public UiTextField()
            : base()
        {
            editor = new TextEditor();
            editor.BeforeNewLine += new EventHandler<TextEditorEventArgs>(editor_BeforeNewLine);

            this.TextAlign = AlignH.Left;
            this.TextValign = AlignV.Middle;
            base.SkinElementName = "Textfield";
            base.TextChanged += new EventHandler(TextField_TextChanged);
        }

        void editor_BeforeNewLine(object sender, TextEditorEventArgs e)
        {
            e.Canceled = !multiline;
        }

        void TextField_TextChanged(object sender, EventArgs e)
        {
            editor.Text = this.text;
        }
        #endregion

        #region Update
        /// <summary>
        /// Updates the textfield.
        /// </summary>
        /// <param name="gameTime"></param>
        /// <param name="buttons"></param>
        /// <param name="position"></param>
        /// <param name="wheelvalue"></param>
        public override void UpdateState(UiManager manager, GameTime gameTime, MouseArgs mouse)
        {
            base.UpdateState(manager, gameTime, mouse);
            this.UpdateBlinkCursor(gameTime);

            if (Focused)
            {
                IInputManager input = manager.Game.Services.GetService(typeof(IInputManager)) as IInputManager;
                editor.Update(input.Keyboard, gameTime.ElapsedGameTime);

                if (editor.Text.Length > 0 && Font != null)
                {
                    Vector2 textSize;
                    int start = 0;
                    int length = editor.CursorPosition;
                    bool measureIsOk;

                    // Measure the text on the left side of the cursor.
                    // start at first character and go upward.
                    do
                    {
                        textSize = Font.MeasureString(editor.Text.Substring(start, length)) * base.ScaleFactorSize;
                        measureIsOk = textSize.X < base.BoundsScaled.Width - (Padding.Left + Padding.Right) * base.ScaleFactorSize.X;
                        if (!measureIsOk)
                        {
                            start++;
                            length--;
                        }
                    } while (!measureIsOk);

                    cursorPosition = Font.MeasureString(editor.Text.Substring(start, length)) * base.ScaleFactorSize + TextPosition;
                    cursorPosition.Y = TextPosition.Y;

                    // Measure the text on the right side of the cursor.
                    // keep the star position of prevous measurement, measure to the end and go downward.
                    length = editor.Text.Length - start;
                    textSize = Font.MeasureString(editor.Text.Substring(start, length)) * base.ScaleFactorSize;
                    while (textSize.X > base.BoundsScaled.Width - (Padding.Left + Padding.Right) * base.ScaleFactorSize.X)
                    {
                        length--;
                        textSize = Font.MeasureString(editor.Text.Substring(start, length)) * base.ScaleFactorSize;
                    }
                    text = editor.Text.Substring(start, length);
                }
                else
                {
                    /*
                     * When the textvalue of the editor has the size 0, then we make a little hack
                     * that makes the cursor stay in the right position.
                     * We apply one character to the text, then the textposition recalculates automaticly.
                     * Then the cursor will be well positioned.
                     */
                    Text = " ";
                    cursorPosition = TextPosition;
                    Text = "";
                    text = "";
                }
            }
            else
            {
                if (editor.Text.Length > 0)
                {
                    Vector2 size = Font.MeasureString(editor.Text) * base.ScaleFactorSize;
                    int start = 0;
                    int length = Text.Length;

                    while (size.X > base.BoundsScaled.Width)
                    {
                        size = Font.MeasureString(editor.Text.Substring(start, length)) * base.ScaleFactorSize;
                        length--;
                    }
                    text = editor.Text.Substring(start, length);
                }
            }
            base.Text = text;
        }

        private void UpdateBlinkCursor(GameTime gameTime)
        {
            cursorBlinkTimer += gameTime.ElapsedGameTime.Milliseconds;
            if (cursorBlinkTimer >= cursorBlinkrate)
            {
                cursorBlinkTimer = 0;
                cursorIsVisible = !cursorIsVisible;
            }
        }
        #endregion

        #region Draw
        /// <summary>
        /// Draws the textfield.
        /// </summary>
        /// <param name="spriteBatch"></param>
        /// <param name="texture"></param>
        public override void Draw(SpriteBatch spriteBatch)
        {
            base.Draw(spriteBatch);

            // Draw the cursor
            if (Font != null && Focused && cursorIsVisible)
            {
                spriteBatch.DrawString(Font, cursor, cursorPosition, TextColor,
                    0, Vector2.Zero, ScaleFactorSize, SpriteEffects.None, 0);
            }
        }

        /// <summary>
        /// Draws the text of the textfield. This is calld inside the Draw Method.
        /// </summary>
        /// <param name="spriteBatch"></param>
        protected override void DrawText(SpriteBatch spriteBatch)
        {
            if (Font != null && text != null)
            {
                spriteBatch.DrawString(Font, text, TextPosition, TextColor,
                    0, Vector2.Zero, ScaleFactorSize, SpriteEffects.None, 0);
            }
        }
        #endregion
    }
}
