﻿using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using CellarDoor.Gui.Elements;
using CellarDoor.Gui.Controller;
using CellarDoor.Gui.Animation;
using CellarDoor.Gui.Theming;
using CellarDoor.Gui.Layout;

namespace CellarDoor.Gui.Layout
{
    public class UiVisualStyle
    {
        /// <summary>
        /// Gets and sets a value indicating whether the theme has to be loaded from manager or not
        /// </summary>
        public bool ThemeInvalid { get; set; }

        /// <summary>
        /// Gets and sets a value indicating whether the state layer hast to be updated.
        /// </summary>
        public bool StateInvalid { get; set; }

        /// <summary>
        /// Gets the element that is styled with this item
        /// </summary>
        public UiElement Element { get; private set; }

        private string elementName;
        /// <summary>
        /// Gets and sets a name to be used to lookup for theme sprites
        /// </summary>
        public string ElementName 
        {
            get { return elementName; }
            set
            {
                if (elementName != value)
                {
                    elementName = value;
                    ThemeInvalid = true;
                }
            }
        }

        /// <summary>
        /// Gets the overlay layer
        /// </summary>
        public UiVisualLayer Overlay { get; protected set; }

        /// <summary>
        /// Gets the state representation layer
        /// </summary>
        public UiVisualLayer StateLayer { get; protected set; }

        /// <summary>
        /// Gets the shadow layer
        /// </summary>
        public UiVisualLayer ShadowLayer { get; protected set; }

        private UiVisualState state;
        /// <summary>
        /// Gets and sets the visual state
        /// </summary>
        public UiVisualState State 
        {
            get { return state; }
            set
            {
                if (state != value)
                {
                    state = value;
                    StateInvalid = true;
                }
            }
        }

        private string fontName;
        /// <summary>
        /// Gets and sets the font name to lookup in the theme definition
        /// </summary>
        public string FontName 
        {
            get 
            {
                if (fontName == null)
                {
                    if (Element.Parent != null)
                    {
                        return Element.Parent.Style.FontName;
                    }
                }
                return fontName ?? "font"; 
            }
            set
            {
                if (fontName != value)
                {
                    fontName = value;
                    ThemeInvalid = true;
                }
            }
        }

        /// <summary>
        /// Gets the used font
        /// </summary>
        public SpriteFont Font { get; protected set; }

        public UiVisualStyle(UiElement element)
        {
            if (element == null)
            {
                throw new ArgumentNullException("element");
            }
            Element = element;

            Overlay = new UiVisualLayer();
            StateLayer = new UiVisualLayer();
            ShadowLayer = new UiVisualLayer();

            Overlay.Enabled = false;
            StateLayer.Enabled = true;
            ShadowLayer.Enabled = false;

            ThemeInvalid = true;
            StateInvalid = true;
        }

        public void Update()
        {
            if (ThemeInvalid)
            {
                UiTheme theme = Element.Manager.Theme;
                int resX = Element.Manager.Viewport.Width;
                int resY = Element.Manager.Viewport.Height;

                LoadTheme(theme, resX, resY);
                ThemeInvalid = false;
            }

            if (StateInvalid)
            {
                UpdateVisualState();
                StateInvalid = false;
            }
        }

        protected virtual void LoadTheme(UiTheme theme, int resX, int resY)
        {
            if (ElementName != null)
            {
                this.Overlay.Sprite = theme.GetState(ElementName, "Overlay", resX, resY);
                this.StateLayer.Sprite = theme.GetState(ElementName, "Default", resX, resY);
                this.ShadowLayer.Sprite = theme.GetState(ElementName, "Shadow", resX, resY);
            }
            else
            {
                this.Overlay.Sprite = null;
                this.StateLayer.Sprite = null;
                this.ShadowLayer.Sprite = null;
            }

            if (FontName != null)
            {
                this.Font = theme.GetFont(FontName, resX, resY);
            }
            else
            {
                this.Font = null;
            }
        }

        protected virtual void UpdateVisualState()
        {
        }

        public virtual void SetTargetArea(Rectangle area)
        {
            Overlay.SetTargetArea(area);
            StateLayer.SetTargetArea(area);
            ShadowLayer.SetTargetArea(area);
        }

        public virtual void Draw(UiManager manager, Color color)
        {
            Draw(manager, ShadowLayer, color);
            Draw(manager, StateLayer, color);
            Draw(manager, Overlay, color);
        }

        protected void Draw(UiManager manager, UiVisualLayer layer, Color color)
        {
            if (layer.Enabled && layer.Sprite != null)
            {
                manager.DrawSprite(layer.Sprite, layer.TargetExtended, color);
            }
        }
    }
}
