#region FileComment
/*
 * Date: 16.03.08
 * Author: Alexander Grfenstein
 * Filename: RectangleFloat.cs
 */
#endregion

#region Using Namespaces
using Microsoft.Xna.Framework;
#endregion

namespace CellarDoor.Gui
{
    /// <summary>
    /// A Rectangle struct that is the same as the Rectangle from the
    /// XNA Framework but uses floatingpoint values.
    /// </summary>
    public struct RectangleF
    {
        #region Variables
        /// <summary>
        /// Specifies the x-coordinate of the rectangle.
        /// </summary>
        public float X;

        /// <summary>
        /// Specifies the y-coordinate of the rectangle.
        /// </summary>
        public float Y;
        
        /// <summary>
        /// Specifies the width of the rectangle.
        /// </summary>
        public float Width;

        /// <summary>
        /// Specifies the height of the rectangle.
        /// </summary>
        public float Height;
        #endregion

        #region Properties
        /// <summary>
        /// Returns an empty rectangle.
        /// </summary>
        public static RectangleF Empty
        {
            get { return new RectangleF(0, 0, 0, 0); }
        }

        /// <summary>
        /// Returns the y-coordinate of the top of the rectangle.
        /// </summary>
        public float Top
        {
            get { return Y; }
        }

        /// <summary>
        /// Returns the y-coordinate of the bottom of the rectangle.
        /// </summary>
        public float Bottom
        {
            get { return Y + Height; }
        }

        /// <summary>
        /// Returns the x-coordinate of the left side of the rectangle.
        /// </summary>
        public float Left
        {
            get { return X; }
        }

        /// <summary>
        /// Returns the x-coordinate of the right side of the rectangle.
        /// </summary>
        public float Right
        {
            get { return X + Width; }
        }
        #endregion

        #region Constructor
        /// <summary>
        /// Creates a rectangle struct that uses floatingpoint values.
        /// </summary>
        /// <param name="X">The x-coordinate of the upper left vertice.</param>
        /// <param name="Y">The y-coordinate of the upper left vertice.</param>
        /// <param name="Width">The width of the rectangle.</param>
        /// <param name="Height">The height of the rectangle.</param>
        public RectangleF(float X, float Y, float Width, float Height)
        {
            this.X = X;
            this.Y = Y;
            this.Width = Width;
            this.Height = Height;
        }

        public RectangleF(Vector2 position, float Width, float Height)
        {
            this.X = position.X;
            this.Y = position.Y;
            this.Width = Width;
            this.Height = Height;
        }

        public RectangleF(float X, float Y, Vector2 size)
        {
            this.X = X;
            this.Y = Y;
            this.Width = size.X;
            this.Height = size.Y;
        }

        public RectangleF(Vector2 position, Vector2 size)
        {
            this.X = position.X;
            this.Y = position.Y;
            this.Width = size.X;
            this.Height = size.Y;
        }

        /// <summary>
        /// Creates a rectangle struct that uses floatingpoint values.
        /// </summary>
        /// <param name="rectangle">Rectangle struct that uses integer values.</param>
        public RectangleF(Rectangle rectangle)
        {
            this.X = rectangle.X;
            this.Y = rectangle.Y;
            this.Width = rectangle.Width;
            this.Height = rectangle.Height;
        }
        #endregion

        #region Methods
        /// <summary>
        /// Determines whether this reactangle entierly contains the specified coordinates.
        /// </summary>
        /// <param name="x">The x-coordinate of the specified point.</param>
        /// <param name="y">The y-coordinate of the specified point.</param>
        /// <returns>True if the recangle contains the specified coordinates.</returns>
        public bool Contains(float x, float y)
        {
            return (x >= X && x <= X + Width &&
                y >= Y && y <= Y + Height);
        }

        /// <summary>
        /// Determines whether this reactangle entierly contains the specified point.
        /// </summary>
        /// <param name="point">The point to evaluate.</param>
        /// <returns>Ture if the rectangle contains the specified point.</returns>
        public bool Contains(Point point)
        {
            return Contains(point.X, point.Y);
        }

        /// <summary>
        /// Determines whether this reactangle entierly contains the specified point.
        /// </summary>
        /// <param name="point">The point to evaluate.</param>
        /// <returns>True if the rectangle contains the specified point</returns>
        public bool Contains(Vector2 point)
        {
            return Contains(point.X, point.Y);
        }

        /// <summary>
        /// Determines whether this reactangle entierly contains the specified rectangle.
        /// </summary>
        /// <param name="rectangle">The rectangle to evaluate.</param>
        /// <returns>True if the rectangle contains the specified rectangle.</returns>
        public bool Contains(RectangleF rectangle)
        {
            return (rectangle.Top >= Top && rectangle.Bottom <= rectangle.Bottom &&
                rectangle.Left >= Left && rectangle.Right <= Right);
        }

        /// <summary>
        /// Determines whether this reactangle entierly contains the specified rectangle.
        /// </summary>
        /// <param name="rectangle">The rectangle to evaluate.</param>
        /// <returns>True if the rectangle contains the specified rectangle.</returns>
        public bool Contains(Rectangle rectangle)
        {
            return (rectangle.Top >= Top && rectangle.Bottom <= rectangle.Bottom &&
                rectangle.Left >= Left && rectangle.Right <= Right);
        }

        /// <summary>
        /// Pushes the edges of the rectangle out by the horizontal and vertical values specified.
        /// </summary>
        /// <param name="horizontalAmoint">Value to push the sides out by</param>
        /// <param name="verticalAmount">Value to push top and bottom out by</param>
        public void Inflate(float horizontalAmoint, float verticalAmount)
        {
            Width += horizontalAmoint;
            Height += verticalAmount;
            X -= horizontalAmoint / 2;
            Y -= verticalAmount / 2;
        }

        /// <summary>
        /// Determines whether a specified rectangle intersects with this rectangle.
        /// </summary>
        /// <param name="rect">The rectangle to evaluate.</param>
        /// <returns>True if specified rectangle intersects with this one; false otherwise.</returns>
        public bool Intersects(RectangleF rect)
        {
            return (this.Contains(rect.X, rect.Y) || this.Contains(rect.Right, rect.Top) ||
                this.Contains(rect.X, rect.Bottom) || this.Contains(rect.Right, rect.Bottom) ||
                rect.Contains(this.X, this.Y) || rect.Contains(this.Right, this.Top) ||
                rect.Contains(this.X, this.Bottom) || rect.Contains(this.Right, this.Bottom));
        }

        /// <summary>
        /// Determines whether a specified rectangle intersects with this rectangle.
        /// </summary>
        /// <param name="rect">The rectangle to evaluate.</param>
        /// <param name="result">True if specified rectangle intersects with this one; false otherwise.</param>   
        public void Intersects(RectangleF rect, out bool result)
        {
            result = (this.Contains(rect.X, rect.Y) || this.Contains(rect.Right, rect.Top) ||
                this.Contains(rect.X, rect.Bottom) || this.Contains(rect.Right, rect.Bottom) ||
                rect.Contains(this.X, this.Y) || rect.Contains(this.Right, this.Top) ||
                rect.Contains(this.X, this.Bottom) || rect.Contains(this.Right, this.Bottom));
        }

        /// <summary>
        /// Changes the position of the rectangle.
        /// </summary>
        /// <param name="x">Change in the x position.</param>
        /// <param name="y">Change in the y position.</param>
        public void Offset(float x, float y)
        {
            X = x;
            Y = y;
        }

        /// <summary>
        /// Changes the position of the rectangle.
        /// </summary>
        /// <param name="amount">The values to adjust the position of the rectangle by.</param>
        public void Offset(Vector2 amount)
        {
            X += amount.X;
            Y += amount.Y;
        }

        /// <summary>
        /// Changes the position of the rectangle.
        /// </summary>
        /// <param name="amount">The values to adjust the position of the rectangle by.</param>
        public void Offset(Point amount)
        {
            X += amount.X;
            Y += amount.X;
        }

        /// <summary>
        /// Converts this rectangle into struct that uses integer values.
        /// </summary>
        /// <returns></returns>
        public Rectangle ToRectangle()
        {
            return new Rectangle((int)X, (int)Y, (int)Width, (int)Height);
        }

        /// <summary>
        /// Retrieves a string representation of the current rectangle.
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "{" + X.ToString() + 
                   " " + Y.ToString() + 
                   " " + Width.ToString() + 
                   " " + Height.ToString() + "}";
        }
        #endregion

        #region StaticMethods
        public static RectangleF IntToFloatRect(Rectangle r)
        {
            return new RectangleF(r.X, r.Y, r.Width, r.Height);
        }
        #endregion
    }
}
