#region FileComment
/*
 * Date: 12.03.08
 * Author: Alexander Grfenstein
 * Filename: RectangleTiled.cs
 */
#endregion

#region Using Namespaces
using System;
using Microsoft.Xna.Framework;
using System.Xml.Serialization;
#endregion

namespace CellarDoor.Gui
{
    /// <summary>
    /// A rectangle that wraps 9 tiles. 
    /// Scaling this rectangle along the x-axis will stretch the top, center and bottom tiles. 
    /// Scaling along the y-axis will stretch only the left, center and tight tile. 
    /// The upper left, upper right lower left and lower right will always keep their initial sizes.
    /// </summary>
    [System.Diagnostics.DebuggerDisplay("{X} {Y} {Width} {Height}")]
    public struct UiRectangle
    {
        #region Variables
        private Rectangle rectTL;
        private Rectangle rectTM;
        private Rectangle rectTR;

        private Rectangle rectL;
        private Rectangle rectM;
        private Rectangle rectR;

        private Rectangle rectBL;
        private Rectangle rectBM;
        private Rectangle rectBR;

        private int x;
        private int y;
        private int width;
        private int height;

        private bool[] empty;
        #endregion

        #region Properties
        /// <summary>
        /// Gets a boolean array where a value indicates if the corresponding tile is empty or not.
        /// </summary>
        public bool[] Empty
        {
            get { return empty; }
        }

        /// <summary>
        /// Gets or sets the the x-coordinate of the rectangle.
        /// </summary>
        public int X
        {
            get { return x; }
            set { SetPosition(value, y); }
        }

        /// <summary>
        /// Gets or sets the y-coordinate of the rectangle.
        /// </summary>
        public int Y
        {
            get { return y; }
            set { SetPosition(x, value); }
        }

        /// <summary>
        ///  Gets or sets the width of the rectangle.
        /// </summary>
        public int Width
        {
            get { return width; }
            set { SetSize(value, height); }
        }

        /// <summary>
        ///  Gets or sets the height of the rectangle.
        /// </summary>
        public int Height
        {
            get { return height; }
            set { SetSize(width, value); }
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Initializes a new instance of the RectangleTiled strict.
        /// </summary>
        /// <param name="TL">A rectangle that represents the top left tile.</param>
        /// <param name="BR">A rectangle that represents the bottom right tile.</param>
        public UiRectangle(Rectangle TL, Rectangle BR)
        {
            this.x = TL.X;
            this.y = TL.Y;
            this.width = BR.Right - TL.Left;
            this.height = BR.Bottom - TL.Top;

            Rectangle TR = new Rectangle(BR.X, TL.Y, BR.Width, TL.Height);
            Rectangle BL = new Rectangle(TL.X, BR.Y, TL.Width, BR.Height);

            // Top Left Tile
            rectTL = TL;
            // Top Right Tile
            rectTR = TR;
            // Bottom Left Tile
            rectBL = BL;
            // Bottom Right Tile
            rectBR = BR;

            // Top Tile
            rectTM = new Rectangle(TL.Right, TL.Y, TR.Left - TL.Right, TL.Height);
            // Left Tile
            rectL = new Rectangle(TL.X, TL.Bottom, TL.Width, BL.Top - TL.Bottom);
            // Right Tile
            rectR = new Rectangle(TR.X, TR.Bottom, TR.Width, BR.Top - TR.Bottom);
            // Bottom Tile
            rectBM = new Rectangle(BL.Right, BL.Y, BR.Left - BL.Right, BL.Height);

            // Middle Tile
            rectM = new Rectangle(rectTM.X, rectL.Y, rectTM.Width, rectL.Height);

            this.empty = new bool[9];
            CheckEmpty();
        }
        #endregion

        #region Methods
        /// <summary>
        /// Chenges the position of the rectangle.
        /// </summary>
        /// <param name="x">Change the x-coordinate.</param>
        /// <param name="y">Change the y-coordinate.</param>
        public void SetPosition(int x, int y)
        {
            this.x = x;
            this.y = y;

            // Top Left
            rectTL.X = x;
            rectTL.Y = y;

            // Top
            rectTM.X = rectTL.Right;
            rectTM.Y = y;
            
            // Top Right
            rectTR.X = rectTM.Right;
            rectTR.Y = y;
            
            // Left
            rectL.X = rectTL.X;
            rectL.Y = rectTL.Bottom;
            
            // Middle
            rectM.X = rectTM.X;
            rectM.Y = rectTM.Bottom;

            // Right
            rectR.X = rectTR.X;
            rectR.Y = rectTR.Bottom;

            // BotLeft
            rectBL.X = rectL.X;
            rectBL.Y = rectL.Bottom;
            
            // Bottom
            rectBM.X = rectM.X;
            rectBM.Y = rectM.Bottom;
            
            // BotRight
            rectBR.X = rectR.X;
            rectBR.Y = rectR.Bottom;
        }

        /// <summary>
        /// Changes the size of the rectangle.
        /// </summary>
        /// <param name="width">The new width value.</param>
        /// <param name="height">The new height value.</param>
        public void SetSize(int width, int height)
        {
            this.width = width;
            this.height = height;

            int vSize = width - (rectTL.Width + rectTR.Width);
            int hSize = height - (rectTL.Height + rectBL.Height);

            rectTM.Width = vSize;
            rectM.Width = vSize;
            rectBM.Width = vSize;

            rectL.Height = hSize;
            rectM.Height = hSize;
            rectR.Height = hSize;

            rectTM.X = rectTL.Right;
            rectM.X = rectL.Right;
            rectBM.X = rectBL.Right;
            
            rectTR.X = rectTM.Right;
            rectR.X = rectM.Right;
            rectBR.X = rectBM.Right;

            rectL.Y = rectTL.Bottom;
            rectM.Y = rectTM.Bottom;
            rectR.Y = rectTR.Bottom;

            rectBL.Y = rectL.Bottom;
            rectBM.Y = rectM.Bottom;
            rectBR.Y = rectR.Bottom;

            CheckEmpty();
        }

        private void CheckEmpty()
        {
            if (this.empty == null)
            {
                this.empty = new bool[9];
            }

            this.empty[0] = rectTL.Width == 0 || rectTL.Height == 0;
            this.empty[1] = rectTM.Width == 0 || rectTM.Height == 0;
            this.empty[2] = rectTR.Width == 0 || rectTR.Height == 0;

            this.empty[3] = rectL.Width == 0 || rectL.Height == 0;
            this.empty[4] = rectM.Width == 0 || rectM.Height == 0;
            this.empty[5] = rectR.Width == 0 || rectR.Height == 0;

            this.empty[6] = rectBL.Width == 0 || rectBL.Height == 0;
            this.empty[7] = rectBM.Width == 0 || rectBM.Height == 0;
            this.empty[8] = rectBR.Width == 0 || rectBR.Height == 0;
        }

        /// <summary>
        /// Returns the current area of the rectangle.
        /// </summary>
        /// <returns></returns>
        public Rectangle GetArea()
        {
            return new Rectangle(x, y, width, height);
        }

        /// <summary>
        /// Changes the whole area of the rectangle.
        /// </summary>
        /// <param name="area">The new area of the rectangle.</param>
        public void SetArea(Rectangle area)
        {
            SetSize(area.Width, area.Height);
            SetPosition(area.X, area.Y);
        }

        public void SetArea(Rectangle area, UiPadding extent)
        {
            SetSize(area.Width + extent.Left + extent.Right, area.Height + extent.Top + extent.Bottom);
            SetPosition(area.X - extent.Left, area.Y - extent.Top);
        }

        /// <summary>
        /// Determines whether this rectangle contains the specified point.
        /// </summary>
        /// <param name="p">The point to evaluate.</param>
        /// <returns>True if this rectangle contains the point</returns>
        public bool Contains(Point p)
        {
            return (
                p.X >= x && p.X <= x + width &&
                p.Y >= y && p.Y <= y + height);
        }
        #endregion

        #region IRectangleTiled Members


        public int Length
        {
            get { return 9; }
        }

        /// <summary>
        /// Gets or sets one of the tiled rectangles
        /// </summary>
        /// <param name="index">The index of the tile to get or set. Must be a value from 0 to 8</param>
        /// <returns></returns>
        public Rectangle this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0:
                        return rectTL;
                    case 1:
                        return rectTM;
                    case 2:
                        return rectTR;
                    case 3:
                        return rectL;
                    case 4:
                        return rectM;
                    case 5:
                        return rectR;
                    case 6:
                        return rectBL;
                    case 7:
                        return rectBM;
                    case 8:
                        return rectBR;
                    default:
                        throw new ArgumentOutOfRangeException();
                }
            }
            set
            {
                switch (index)
                {
                    case 0:
                        rectTL = value;
                        break;
                    case 1:
                        rectTM = value;
                        break;
                    case 2:
                        rectTR = value;
                        break;
                    case 3:
                        rectL = value;
                        break;
                    case 4:
                        rectM = value;
                        break;
                    case 5:
                        rectR = value;
                        break;
                    case 6:
                        rectBL = value;
                        break;
                    case 7:
                        rectBM = value;
                        break;
                    case 8:
                        rectBR = value;
                        break;
                    default:
                        throw new ArgumentOutOfRangeException();
                }
                CheckEmpty();
            }
        }

        #endregion

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return String.Format("{{0},{1},{2},{3}}", x, y, width, height);
        }

        public static bool operator ==(UiRectangle a, UiRectangle b)
        {
            for (int i = 0; i < a.Length; i++)
            {
                if (a[i] != b[i])
                {
                    return false;
                }
            }
            return true;
        }

        public static bool operator !=(UiRectangle a, UiRectangle b)
        {
            return !(a == b);
        }
    }
}
